# version 2022.1
[CmdLetBinding()]
param(
	[string]$username,
	[string]$password,
	[Switch]$logToConsole
)

try {
	cd $PSScriptRoot
	$logFileName = 'ownership_maintenance.{0}.log' -f (get-date -Format 'yyyy-MM')
	. ./utils.ps1 -logFileName $logFileName -logToConsole:$logToConsole
	$api = [mm]::getAPI($username, $password)
} catch {
	Write-Output ('Got exception while initializing: ' + $_.Exception.GetBaseException().Message)
	exit 1
}

. ./workflow_common.ps1

$subjects = $Global:mmConfig.subjects


function augmentRecords([hashtable[]]$records,[hashtable]$zone){
    foreach ($record in $records){
        $record.zone = $zone
        $record.fqdn = if ($record.name.EndsWith('.')) { $record.name } else { "$($record.name).$($zone.name)" }

        $props = [mm]::getPropertyMap($record.customProperties)
        $record.Owner = $props.Owner
        $record.Manager = $props.Manager
        $record.TeamDL = $props.TeamDL
    }
    return $records
}

function updateManagerIfNeeded([string]$owner, [hashtable[]]$records){
    logv2 "updating manager if needed for $owner"
    $manager = [adUtils]::getManagerForUser($owner).name
    if ([mm]::isEmpty($manager)){
        log "Empty manager for Owner $owner. Has the account been disabled? Manager not updated"
        return
    }

    foreach ($rec in $records){
        if ($rec.Manager -ne $manager){
            try {
                logv2 "updating manager for $owner on records $($rec.fqdn)"
                $api.setProperties($rec.ref, 'DNSRecord', @{Manager=$manager}, "Ownership maintenance process")
            } catch {
                logException 'while setting record properties'
            }
        }
    }
}

function handleStaleOwner([string]$owner,[hashtable[]]$records){
    # we've already determined Owner is stale, cannot be found in AD. We need to ask Team(s) to update ownership via email
    log "Found stale owner $owner. Notifying team(s)"
    $subject =  (?? $subjects['ownership-stale-owner'] "Owner for DNS records needs to be updated")
    $file = 'ownership-stale-owner.htm'
    $zoneID = [mm]::ref2ID($records[0].dnsZoneRef)
    $manager = $records[0].Manager # we can assume the records all have the same manager, since it's the same Owner on all of them
    $recordsByTeam = $records | groupBy 'TeamDL'
    foreach ($team in $recordsByTeam.Keys){
        try {
            $teamMail = [adUtils]::getGroup($team).mail
            $managerMail = if ([mm]::notEmpty($manager)) { [adUtils]::getUserByName($manager).mail } else { $null }
            if ([mm]::isEmpty($teamMail)){
                if ([mm]::isEmpty($managerMail)){
                    log "Neither manager $manager nor $team found in AD or have no email. Unable to send ownership update request notification"
                    continue
                }

                log "Could not inform Team to update owner: not found or no email configured. Sending manager only"
                $teamMail = $managerMail
                $managerMail = $null # no reason to have manager on cc too
            }
            $teamRecs = $recordsByTeam[$team]

            $tokens = @{
                owner = $owner
                team = $team
                filter = "Owner=$($owner.Replace('\','\\'))" # domain\user needs to be filtered as domain\\user
                records=[htmlUtils]::getRecordListAsHTMLListItems($teamRecs)
                zoneID = $zoneID
            }
            [mmEmail]::prepareAndSend($teamMail, $subject, $file, $managerMail, $Global:mmConfig.email.bcc, $tokens)
        } catch {
            logException 'while sending stale owner notification'
        }
    }
}

function handleStaleTeam([string]$team, [string]$reason, [hashtable[]]$records){
    log "Sending stale team notification for team $team because of $reason"
    # we want to ask the Owner to update the TeamDL on the records on input for the given reason
    $subject =  (?? $subjects['ownership-stale-team'] "TeamDL for DNS records needs to be updated")
    $file = 'ownership-stale-team.htm'
    $zoneID = [mm]::ref2ID($records[0].dnsZoneRef)
    $recordsByOwner = $records | groupBy 'Owner'
    foreach ($owner in $recordsByOwner.Keys){
        try {
            $ownerRecs = $recordsByOwner[$owner]

            $ownerMail = [adUtils]::getUserByName($owner).mail
            if ([mm]::isEmpty($ownerMail)){
                # owner orphaned too - the manager would already have received a notification for orphaned owner, so no need to send again for orphaned team
                continue
            }

            $managerMail = [adUtils]::getManagerForUser($owner).mail

            $tokens = @{
                owner = $owner
                team = $team
                filter = "TeamDL=$($team.Replace('\','\\'))" # domain\team needs to be filtered as domain\\team
                records=[htmlUtils]::getRecordListAsHTMLListItems($ownerRecs)
                zoneID = $zoneID
                reason = $reason
            }
            [mmEmail]::prepareAndSend($ownerMail, $subject, $file, $managerMail, $Global:mmConfig.email.bcc, $tokens)
        } catch {
            logException 'while sending email about stale team'
        }
    }
}


# This maintenance involves:
# - verifying owner
#    - if owner is valid, update manager
#    - if owner is not valid send team(s) email to update Owner
#  - verify teamDL
#    - if invalid ask Owner to update TeamDL on records
#    - if valid, but has no members, ask owner to update TeamDL
#    - if valid, but Owner is no longer in Team, ask Owner to update

$zones = $api.call('GetDNSZones', @{filter='type=Master Workflow!=""'}).dnsZones # we'll maintain this on all workflow enabled zones, attestation or not
log "Found $($zones.Count) workflow enabled zones to maintain ownership on"

foreach ($zone in $zones){
    try {
        log "processing zone $($zone.name)"
        $filter = 'Owner!=""' # just process all records that have owner defined - otherwise the record is deemed not to have ownership
        $records = $api.call('GetDNSRecords',@{filter=$filter; dnsZoneRef=$zone.ref}).dnsRecords
        $records = @(augmentRecords -records $records -zone $zone)

        # first, verify owner and update manager
        $recordsByOwner = $records | groupBy 'Owner'
        foreach ($owner in $recordsByOwner.Keys){
            if ([adUtils]::verifyUserByName($owner)){
                updateManagerIfNeeded $owner $recordsByOwner[$owner]
            } else {
                handleStaleOwner $owner $recordsByOwner[$owner]
            }
        }

        # verify team, and notify if stale, empty or not a member
        $recordsByTeam = $records | groupBy 'TeamDL'
        foreach ($team in $recordsByTeam.Keys){
            if ([mm]::isEmpty($team)){
                # no team, no to do
                continue
            }
            $groupInfo = [adUtils]::getGroup($team)
            if (!$groupInfo){
                handleStaleTeam $team "TeamDL `"$team`" not found in Active Directory" $recordsByTeam[$team]
            } elseif ([adUtils]::getUsersInGroupByDN($groupInfo.dn, @(), $true).Count -eq 0) {
                handleStaleTeam $team "TeamDL `"$team`"has no members in Active Directory" $recordsByTeam[$team]
            } else {
                $teamRecsByOwner = $recordsByTeam[$team] | groupBy 'Owner'
                foreach ($owner in $teamRecsByOwner.Keys) {
                    if (![adUtils]::verifyUserByName($owner)){
                        # stale owner, but group exists. We've already dealt with this record in handleStaleOwner above, so let's hop over it and do nothing
                    } elseif (![adUtils]::isUserMemberInGroupByDN($owner, $groupInfo.dn)) {
                        # owner is no longer a member in team registered on these records
                        handleStaleTeam $team "You are not a member of the TeamDL $team in AD (anymore)" $teamRecsByOwner[$owner]
                    }
                }
            }
        }
    } catch {
        logException "while processing zone $($zone.name)"
    }
}

$api.logout()
